@@ -19,6 +19,7 @@ See http://pypi.python.org/pypi/injector for documentation.
 import itertools
 import functools
 import inspect
+import sys
 import types
 import threading
 from abc import ABCMeta, abstractmethod
@@ -26,7 +27,7 @@ from collections import namedtuple
 
 
 __author__ = 'Alec Thomas <alec@swapoff.org>'
-__version__ = '0.6.0'
+__version__ = '0.6.1'
 __version_tag__ = ''
 
 
@@ -42,6 +43,19 @@ def synchronized(lock):
 lock = threading.RLock()
 
 
+def reraise(original, exception):
+    prev_cls, prev, tb = sys.exc_info()
+    frames = inspect.getinnerframes(tb)
+    if len(frames) > 1:
+        exception = original
+    try:
+        raise exception.with_traceback(tb)
+    except AttributeError:
+        # This syntax is not a valid Python 3 syntax so we have
+        # to work around that
+        exec('raise exception.__class__, exception, tb')
+
+
 class Error(Exception):
     """Base exception."""
 
@@ -62,6 +76,9 @@ class CallError(Error):
     """Call to callable object fails."""
 
     def __str__(self):
+        if len(self.args) == 1:
+            return self.args[0]
+
         instance, method, args, kwargs, original_error = self.args
         if hasattr(method, 'im_class'):
             instance = method.__self__
@@ -531,11 +548,12 @@ class Injector(object):
         except TypeError as e:
             # The reason why getattr() fallback is used here is that
             # __init__.__func__ apparently doesn't exist for Key-type objects
-            raise CallError(
+            reraise(e, CallError(
                 instance,
                 getattr(instance.__init__, '__func__', instance.__init__),
                 (), additional_kwargs, e,
                 )
+            )
         return instance
 
     def install_into(self, instance):
@@ -694,7 +712,7 @@ def inject(**bindings):
                     try:
                         return f(self_, *args, **kwargs)
                     except TypeError as e:
-                        raise CallError(self_, f, args, kwargs, e)
+                        reraise(e, CallError(self_, f, args, kwargs, e))
                 dependencies = injector.args_to_inject(
                     function=f,
                     bindings=bindings,
@@ -704,7 +722,7 @@ def inject(**bindings):
                 try:
                     return f(self_, *args, **dependencies)
                 except TypeError as e:
-                    raise CallError(self_, f, args, dependencies, e)
+                    reraise(e, CallError(self_, f, args, dependencies, e))
             # Propagate @provides bindings to wrapper function
             if hasattr(f, '__binding__'):
                 inject.__binding__ = f.__binding__
@@ -716,16 +734,21 @@ def inject(**bindings):
     def class_wrapper(cls):
         orig_init = cls.__init__
 
+        original_keys = tuple(bindings.keys())
+
+        for k in bindings:
+            bindings[k.lstrip('_')] = bindings.pop(k)
+
         @inject(**bindings)
         def init(self, *args, **kwargs):
-            for key in bindings:
-                setattr(self, key, kwargs.pop(key))
+            for key in original_keys:
+                setattr(self, key, kwargs.pop(key.lstrip('_')))
             orig_init(self, *args, **kwargs)
         cls.__init__ = init
         return cls
 
     def multi_wrapper(something):
-        if type(something) is type:
+        if isinstance(something, type):
             return class_wrapper(something)
         else:
             return method_wrapper(something)
